/* * (C) Copyright IBM Corporation 2016. */
/*
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA


 Author: IBM Corporation
 This module is part of the IBM (R) Rational (R) ClearCase (R)
 Multi-version file system (MVFS).
 For support, please visit http://www.ibm.com/software/support

*/
#ifndef _MVFS_LPC_LINUX_H_
#define _MVFS_LPC_LINUX_H_
#include <linux/types.h>        /* We need uint64_t from here. */
#include <linux/wait.h>         /* Where waitqueues are defined. */
#include <linux/jiffies.h>      /* For timeout conversions. */
#include <linux/sunrpc/msg_prot.h> /* For enum rpc_accept_stat. */

/* Use a macro for logging so we can centralize the use of the MFS_VIOCTL flag
** (defined as 0x00000008 in sys/mfs_stats.h) to control our logging as well as
** do things differently for MVFS_DEBUG.  Use MVFS_LPC_LL as the Log Level.
*/
#if defined(MVFS_DEBUG)
#define MVFS_LPC_LL MFS_LOG_ERR
#else
#define MVFS_LPC_LL MFS_LOG_DEBUG
#endif

/* Note, ", ##__VA_ARGS__" elides that comma if __VA_ARGS__ is empty. */
#define MVFS_LPC_DBG(format, ...) \
    if (MDB_IS_VLOG_SET(MFS_VIOCTL)) \
        mvfs_log(MVFS_LPC_LL, "%s: " format, __func__, ##__VA_ARGS__)

/* The structure added to the pending or sent list in ic_linux_h_t (below).  It
** contains the buffer arguments that were passed to mvfs_lpc_linux_call (so
** that the IOCTL side can do copyout of the kernel request and copyin of the
** user response), a handle that we will pass to user space so when the response
** comes back with the handle we can find the matching request, and a waitqueue
** that this kernel request will wait on for the response (or a timeout).
*/
typedef enum ic_krn_req_type {
    IC_UNSET,
    IC_PENDING,
    IC_SENT
} ic_krn_req_type_t;

typedef struct ic_krn_req {
    struct ic_krn_req *ickr_next;    /* Pending or sent list link. */
    ic_krn_req_type_t ickr_type;     /* Which list type. */
    uint64_t ickr_handle;            /* User space handle. */
    wait_queue_head_t ickr_waitqueue;
    tbs_boolean_t ickr_res_available;
    enum clnt_stat ickr_res_stat;
    int ickr_op;
    char *ickr_input_bufp;
    size_t ickr_input_bufsz;
    char *ickr_output_bufp;
    size_t ickr_output_bufsz;
} ic_krn_req_t;

/* Our handle type, which includes the waitqueue where IOCTL calls can wait for
** a kernel RPC request and linked lists of pending and sent kernel RPC
** requests.  Note, an IOCTL call will only add itself to the waitqueue if the
** pending list is empty.  The init_waitqueue_head(&(h.ic_ioctl_waitqueue) call
** is made when the handle is created in mvfs_lpc_create.  The pending list is
** managed FIFO, added to the tail and removed from the head, so no request
** should get "starved" (which it might if we managed the list LIFO or some
** other way).  We use the xid set by the RPC call as the user space handle.  In
** order to validate the handle when it comes back we keep a sent list (LIFO
** because it's easier, but order doesn't matter).  If the handle isn't on that
** list, it must be invalid and we can just drop the response.  The handle lock
** covers both lists as well as the setting of the response fields in the
** request structure and copyin/copyout and wakeups.
**
** Some design notes: If we could trust user space (and didn't mind showing
** kernel pointers to user space), we could have made the handle be the kernel
** pointer to the request structure and then we wouldn't need a sent list since
** we would just use the handle (as a pointer) to find the request.  Also note,
** these lists are searched linearly since we don't expect very many pending or
** sent requests (the rview agent probably only has 5 or 10 threads, so that's
** the largest number of requests that could be on the sent list).
*/
typedef struct ic_linux_h {
    LOCK_T ic_h_lock;
    tbs_boolean_t ic_shutting_down;
    wait_queue_head_t ic_ioctl_waitqueue;
    ic_krn_req_t *ic_pend_head;
    ic_krn_req_t *ic_pend_tail;
    ic_krn_req_t *ic_sent_head;
} ic_linux_h_t;
#define MVFS_LPC_MDEP_H_T ic_linux_h_t *

extern int mvfs_lpc_linux_teardown(void);
#define MVFS_LPC_MDEP_TEARDOWN mvfs_lpc_linux_teardown

extern int mvfs_lpc_linux_initialize(void);
#define MVFS_LPC_MDEP_INITIALIZE mvfs_lpc_linux_initialize

extern int mvfs_lpc_linux_destroy(MVFS_LPC_MDEP_H_T h);
#define MVFS_LPC_MDEP_DESTROY(h) mvfs_lpc_linux_destroy(h)

extern int
mvfs_lpc_linux_create(
    const char *nm,
    CRED_T *cred,
    MVFS_LPC_MDEP_H_T *hpp
);
#define MVFS_LPC_MDEP_CREATE(nm, cred, hp) \
    mvfs_lpc_linux_create((nm), (cred), (hp))

extern int mvfs_lpc_linux_flush(MVFS_LPC_MDEP_H_T h);
#define MVFS_LPC_MDEP_FLUSH(h) mvfs_lpc_linux_flush(h)

enum clnt_stat
mvfs_lpc_linux_call(
    MVFS_LPC_MDEP_H_T h,
    int op,
    XID_T xid,
    u_long timeout,
    char *input_buf,
    size_t input_buf_size,
    char *output_buf,
    size_t output_buf_size
);
#define MVFS_LPC_MDEP_CALL(h, \
                           op, \
                           xid, \
                           timeout, \
                           input_buffer, \
                           input_buffer_size, \
                           output_buffer, \
                           output_buffer_size) \
    mvfs_lpc_linux_call((h), \
                        (op), \
                        (xid), \
                        (timeout), \
                        (input_buffer), \
                        (input_buffer_size), \
                        (output_buffer), \
                        (output_buffer_size))

int
mvfs_lpc_linux_ioctl(
    MVFS_LPC_MDEP_H_T h,
    void *datap,
    u_long datalen,
    CALL_DATA_T *cd,
    MVFS_CALLER_INFO *callinfo
);
#define MVFS_LPC_MDEP_IOCTL(h, \
                            datap, \
                            datalen, \
                            cd, \
                            callinfo) \
    mvfs_lpc_linux_ioctl(h, \
                         datap, \
                         datalen, \
                         cd, \
                         callinfo)

#endif /* _MVFS_LPC_LINUX_H_ */
/* $Id: ba88d9d2.51d711e7.871d.44:37:e6:65:93:e5 $ */
